package top.ninwoo.test.doMap;

import java.util.Arrays;
import java.util.Random;


/**
 * 求某一数据量下的最佳映射结果和业务处理时延
 * 输出最佳粒子的位置和适应度值
 * 目前只允许计算过程跨单时隙
 * @jthu
 */
public class DoMap {
	

    //最大权值
    static final int MAXWEIGHT = 10000*1000;

    /*
     * 业务参数
     */
    //数据量MB？
    static double D = 5;

    //子任务数
    static int c = 8;

    //子任务依赖关系
    static int[][] link = {{0, 1, 1, 2, 3, 4, 5, 6}, {1, 2, 3, 4, 5, 6, 6, 7}};

    //固定映射任务数==起始任务+终止任务
    static final  int C_STABLE = 2;

    //实际需映射任务数
    static int n1 = c - C_STABLE;

    /*
     * 网络参数
     */
    //设备数
    static int num = 20;
    //设备计算能力参数s  GB/s
    static double[] nt = {1.112341341313, 0.5196272109574349, 0.675827555541419, 0.7637869915859579, 0.7998320296858645,
            0.638652700648973, 0.6396760676518485, 0.5615337810185533, 0.6356797503522186, 0.8266103444996542, 0.6924037701405794,
            0.9371892689614395, 0.6949256728585527, 0.6212614126622614, 0.6318393608430264, 0.7494680794385384, 0.9983949989039682,
            0.5610722256465777, 0.6415780028672076, 0.9371892689};  //20节点
    //时隙大小
    static double slot = 20*1000;
    //时隙数目
    static int slotnum;
    //网络节点数
    static int dev;
    //初始化时延偏移量
    //这里需要一个时钟输入
    static double testLatency = 0;

    //固定映射设备数==起始设备+终止设备
    static final  int N_STABLE = 2;
    //实际可映射设备数
    static int n2;
    //单时隙可映射设备数
    static int mapdev = num - N_STABLE ;
    /*
     * BPSO参数
     */
    //粒子群规模
    static int n = 200;
    //种群迭代次数
    static int k = 500;
    //惯性权重，加速因子1，2
    static double s = 1.5; static int c1 = 1; static int c2 = 1;

    public static void main(String[] args) {
        doMapApp();
    }

    public static void doMapApp() {


        for (slotnum = 1;slotnum<=3;slotnum++){
            dev = num*slotnum;
            n2 = dev - N_STABLE*slotnum;
            //组装最短路矩阵
            double[][] edgeWeight = WTEG();
            InterSlotDijstraV2 dijs = new InterSlotDijstraV2(num, slot);
            InterSlotDijstraV2.pathRecord pathRecord = dijs.dijstra(edgeWeight, testLatency);
            double[][] lt = pathRecord.pathWeight;
            for (int i = 0;i<nt.length;i++) {
                nt[i] = (D * 1000) / nt[i];
            }
            //执行粒子群算法，得路径计算结果
            DoMap doMap = new DoMap();
            Par best_par = doMap.main_bpso(k, c, dev, n1, n2, D, lt, nt, link, n, s, c1, c2, num, slot, mapdev);
            if (best_par.fit != 0){
                System.out.println(1 / best_par.fit);
                System.out.println(Arrays.toString(best_par.task));
                System.out.println(Arrays.toString(best_par.taskfinal));
                break;
            }
            //根据分配结果，回溯每一条路由

        }
    }

    private static double[][] WTEG() {
        //获得单时隙连接矩阵  0或MAXWEIGHT
        double[][] slot1 = new double[num][num];
        for(int i = 0;i<slot1.length;i++){
            for(int j = 0;j<slot1.length;j++){
                slot1[i][j] = 0;
            }
        }
        double[][] slot2 = new double[num][num];
        for(int i = 0;i<slot2.length;i++){
            for(int j = 0;j<slot2.length;j++){
                slot2[i][j] = 0;
            }
        }
        double[][] slot3 = new double[num][num];
        for(int i = 0;i<slot3.length;i++){
            for(int j = 0;j<slot3.length;j++){
                slot3[i][j] = 0;
            }
        }
        //生成相邻时隙矩阵
        double[][] neighbor = new double[num][num];
        for(int i = 0;i<neighbor.length;i++){
            for(int j = 0;j<neighbor.length;j++){
                if(i == j){
                    neighbor[i][j] = slot;
                }else{
                    neighbor[i][j] = MAXWEIGHT;
                }
            }
        }
        //生成其余时隙矩阵
        double[][] nonNeighbor = new double[num][num];
        for(int i = 0;i<nonNeighbor.length;i++){
            for(int j = 0;j<nonNeighbor.length;j++){
                nonNeighbor[i][j] = MAXWEIGHT;
            }
        }
        //组装连接矩阵
        double[][] edgeWeight = new double[dev][dev];
        for(int i = 0;i<edgeWeight.length;i++){
            for(int j = 0;j<edgeWeight.length;j++){
                if (i>=0 & i<num & j>=0 & j<num) {
                    edgeWeight[i][j] = slot1[i][j];
                }else if (i>=num & i<2*num & j>=num & j<2*num) {
                    edgeWeight[i][j] = slot2[i%num][j%num];
                }else if (i>=2*num & i<3*num & j>=2*num & j<3*num) {
                    edgeWeight[i][j] = slot3[i%num][j%num];
                }else if ((i>=0 & i<num & j>=num & j<2*num) || (i>=num & i<2*num & j>=2*num & j<3*num)){
                    edgeWeight[i][j] = neighbor[i%num][j%num];
                }else {
                    edgeWeight[i][j] = nonNeighbor[i%num][j%num];
                }

            }
        }
        return edgeWeight;
    }


    /*
     * 粒子群算法主方法
     */
    private Par main_bpso(int k, int c, int dev, int n1, int n2, 
    		double D, double[][] lt, double[] nt, int[][] link, int n,double s, int c1, int c2, int num, double slot, int mapdev) {
    	//初始化粒子群数组
    	int nn = n1*n2;  //粒子维度
        Par best_par = new Par(nn, c);
        Par[] par_array = new Par[n];
    	Par [] temp = new Par[2];			//存放返回值
        for (int i = 0; i < par_array.length; i++) {
        	par_array[i] = new Par(nn, c);
			initialize(par_array[i],n1,n2, c, dev, num, mapdev);
			
        //计算粒子适应度值
            par_array[i].fit = compute_fit(par_array[i], D, lt, nt, c, dev, n1, n2, nn, link, num, slot);
            
        //更新群体最优粒子
//        	Par [] temp = new Par[2];			//存放返回值
        	temp = swarmBest(par_array[i], best_par, nn, c);
        	par_array[i] = temp[0];
        	best_par = temp[1];
        }

        /*
         * 更新粒子群
         */
        for (int i = 0; i < k; i++) {			//迭代次数
            for (int j = 0; j < n; j++) {		//种群规模
            	
            	//更新粒子位置、速度
                update_par(par_array[j], best_par, nn, s, c1, c2, c, n1, n2, dev, num, mapdev);
                
                //计算粒子适应度值
                par_array[j].fit = compute_fit(par_array[j], D, lt, nt, c, dev, n1, n2, nn, link, num, slot);
                
                //更新群体最优粒子
            	temp = swarmBest(par_array[j], best_par,nn, c);
            	par_array[j] = temp[0];
            	best_par = temp[1];
            }
        }
        return best_par;
    }

    /*
     * 初始化粒子位置、速度
     */
    private void initialize(Par par, int n1, int n2, int c, int dev, int num, int mapdev) {
    	Random random = new Random();
    	//初始化位置
    	for (int i = 0; i < par.x.length; i+= n2) {
    		 int rand = random.nextInt(n2);       //功能：随机生成1 -- n2的任意 int 值
    		 par.x[i+rand] = 1;
		}
    	
        //extract 提取映射结果
        par.task[0] = 0;
        par.task[c - 1] = dev - 1;
        int index = 1;
        for (int i = 0; i < n1 * n2; i += n2) {
            for (int j = 0; j < n2; j++) {
                if (par.x[i + j] == 1) {
                	par.task[index] = num*(j/mapdev) +(j%mapdev) + 1;  //
                    index++;
                    break;
                }
            }
        }
        
        //排序
        int temp;
        for(int p = 0; p<(c-1); p++) {
        	for(int j = 0; j <(c-1-p); j++) {
        		if(par.task[j] > par.task[j + 1]) {
        			temp = par.task[j + 1];
        			par.task[j + 1] = par.task[j];
        			par.task[j] = temp;
        		}
        	}
        }
        
        for(int i = 0; i < c; i++) {
        	par.taskfinal[i] = par.task[i];
        	if(par.taskfinal[i]>dev){
                System.out.println(286);
            }
        }
    	
//        //extract 提取映射结果
//        par.taskfinal[0] = 0;
//        par.taskfinal[c - 1] = dev - 1;
//        index = 1;
//        for (int i = 0; i < n1 * n2; i += n2) {
//            for (int j = 0; j < n2; j++) {
//                if (par.x[i * n2 + j] == 1) {
//                	par.taskfinal[index] =i * n2 + j + 1;
//                    index++;
//                    break;
//                }
//            }
//        }
    	
    	//初始化速度
    	for (int i = 0; i < par.v.length; i++) {
    		double value = random.nextDouble();
            par.v[i] = value;
		}
	}
    
    /*
     * 更新群体最优粒子
     */
    private Par[] swarmBest(Par par, Par bestPar, int nn, int c) {
    	Par[] temp = new Par[2];
    	//更新局部最优
    	if (par.fit > par.best_fit) {
    		par.best_fit = par.fit;
            for (int j = 0; j <  nn; j++) {
            	par.bestx[j] = par.x[j];
            }
        }
    	temp[0] = par;
    	//更新全局最优
        if (par.best_fit > bestPar.fit) {
        	bestPar.fit = par.best_fit;
            for (int j = 0; j < nn; j++) {
            	bestPar.x[j] = par.bestx[j];
            }
            for(int k =0; k<c; k++) {
            	bestPar.task[k] = par.task[k];
            	bestPar.taskfinal[k] = par.taskfinal[k];
            }
        }
        temp[1] = bestPar;
        
        return temp;
	}
    
    /*
     * 更新粒子位置、速度
     */
    private void update_par(Par par, Par best_par, int nn, double s, int c1, int c2, int c,
    		int n1, int n2, int dev, int num, int mapdev) {
        Random random = new Random();
        /**
         * 公式按公式更新粒子后如果不符合映射规则要求，则再次更新
         * 为此设计布尔变量flag & flag2
         * flag控制是否进行新一轮粒子更新
         * flag2控制是否进行粒子后续映射位置检查
         * 判断流程为 ：初始化flag-->更新粒子一次-->初始化flag2-->判断粒子是否合法-->合法则置flag为false----------------------------
         * 																																		-->非法则置flag2为false,并break内部for循环
         * --------------------------------------------------------------------->完成粒子更新 
         * -->并且continue本次外部for循环-->再重新更新粒子一次
         */
//        boolean flag = true;
//        while (flag) {
        for (int i = 0; i < nn; i++) {
            //更新速度
            par.v[i] = s * par.v[i] + c1 * random.nextDouble() * (best_par.bestx[i] - par.x[i]) + c2 * random.nextDouble() * (par.bestx[i] - par.x[i]);
            if (par.v[i] > 1) {
                par.v[i] = 1;
            }
            if (par.v[i] < -1) {
                par.v[i] = -1;
            }
        }
        //更新位置
        int maxIndex=0;//假设第一个元素为最大值 那么下标设为0
        for(int i =0;i<par.v.length-1;i++){
            if(par.v[maxIndex]<par.v[i+1]){
                maxIndex=i+1;
            }
        }
        for (int i = (maxIndex/n2)*n2;i<(maxIndex/n2+1)*n2;i++){
            if (i == maxIndex){
                par.x[i] = 1;
            }else{
                par.x[i] = 0;
            }
        }

/*            boolean flag2 = true;
        for (int i = 0; i < nn; i += n2) {
            int sum = 0;
            for (int j = 0; j < n2; j++) {
                sum += par.x[i + j];
            }
            if (sum != 1) {
                flag2 = false;
                break;
            }
        }
        if (!flag2) {
            continue;
        }*/
//            flag = false;

        //extract 提取映射结果
        par.task[0] = 0;
        par.task[c - 1] = dev - 1;
        int index = 1;
        for (int i = 0; i < n1 * n2; i += n2) {
            for (int j = 0; j < n2; j++) {
                if (par.x[i + j] == 1) {
                    par.task[index] = num*(j/mapdev) +(j%mapdev) + 1;
                    index++;
                    break;
                }
            }
        }
        //排序
        int temp;
        for(int p = 0; p<(c-1); p++) {
            for(int j = 0; j <(c-1-p); j++) {
                if(par.task[j] > par.task[j + 1]) {
                    temp = par.task[j + 1];
                    par.task[j + 1] = par.task[j];
                    par.task[j] = temp;
                }
            }
        }

        for(int i = 0; i < c; i++) {
            par.taskfinal[i] = par.task[i];
            if(par.taskfinal[i]>dev){
                System.out.println(421);
            }
        }

    }
//    }
    /**
     * 计算适应度值
     */
    private double compute_fit(Par par, double D, double[][] lt, 
    		double[] nt, int c, int dev, int n1, int n2, int nn, int[][] link, int num, double slot) {
    	//
        double[] compu = new double[c];
        //
        double[] accu = new double[c];
//        //初始化任务数组
//        int[] task = new int[c];
//        
//        //extract 提取映射结果
//        task[0] = 0;
//        task[c - 1] = dev - 1;
//        int index = 1;
//        for (int i = 0; i < n1 * n2; i += n2) {
//            for (int j = 0; j < n2; j++) {
//                if (par.x[i + j] == 1) {
//                    task[index] = j;
//                    index++;
//                    break;
//                }
//            }
//        }
//        System.out.println(Arrays.toString(task));
//        //初始化任务数组
//        task = new int[] {0, 7, 4, 4, 7, 7, 11};
        
		/*
		 * 业务处理时延 = 业务累计时延 + 业务计算时延
		 * 业务累计时延 = 前项业务处理时延 + 业务传输时延
		 * 
		 * 业务处理时延 = 前项业务处理时延 +（业务传输时延 + 业务计算时延）
		 */
        compu[0] = 0;
        accu[0] = 0 + compu[0];
        
        for (int i = 1; i < c; i++) {
            compu[i] = nt[par.taskfinal[i]%num];
        }

        //子任务终点
        for (int i = 1; i < c; i++) {
            double max_trans = 0;
            int link_size = link[0].length;
            //子任务起点
            for (int j = 0; j < link_size; j++) {
//                int end = link[1][j];
            	int end = par.taskfinal[link[1][j]];
                if ( link[1][j] == i) {
//                    int start = link[0][j];
                    int start = par.taskfinal[link[0][j]];
                    double transv;
                    if (start/num == end/num){
                        transv = accu[link[0][j]] + lt[start][end];
                    }else{
                        transv = lt[start][end]+start/num*slot;  //传输跨时隙的话，时延应该是跨时隙后的时刻
                    }
                    max_trans = Math.max(max_trans, transv);
                }
            }
            accu[i] = compu[i] + max_trans;
            //modify
            if(accu[i] > (par.taskfinal[i]/num + 1)*slot) {
                if(par.taskfinal[i] + num < dev){
                    par.taskfinal[i] += num;			//缺少参数 num, slot
                    for(int j = i + 1; j<c ;j++) {
                        if(par.taskfinal[j]/num < par.taskfinal[i]/num) {
                            par.taskfinal[j] += num;
                        }
                    }
                }else {
//                    System.out.println("本方案不可行");
                    return 0;
                }
            }
        }
        double fit = 1 / accu[c - 1];
        return fit;
    }

    /**
     * 粒子引用类型，是粒子群算法搜索的最小单元
     *  输入参数：位置（自变量）、速度、适应度、[个体最优位置、个体最优适应度(解)]
     */
    class Par {
    	//粒子群输入参数声明
        int[] x;
        double[] v;
        double fit;
        double best_fit;
        int[] bestx;
        int[] task;
        int[] taskfinal;
        
        //构造函数
        public Par(int nn,int c) {
        	x = new int[nn];
        	v = new double[nn];
        	bestx = new int[nn];
        	fit = 0;
        	best_fit = 0;
        	task = new int[c];
        	taskfinal = new int[c];
		}
        
    }

}